from __future__ import annotations
from enum import Enum
import re
from typing import Callable, Never

import typer

from reling.types import Promise
from reling.utils.functions import named_function
from reling.utils.strings import replace_prefix_casing

__all__ = [
    'TyperExtraOption',
    'typer_enum_autocompletion',
    'typer_enum_options',
    'typer_enum_parser',
    'typer_func_parser',
    'typer_raise',
    'typer_raise_import',
    'typer_regex_parser',
]


def typer_raise(message: str) -> Never:
    typer.echo(message, err=True)
    raise typer.Exit(code=1)


def typer_raise_import(library: str) -> Never:
    typer_raise(f'{library} could not be imported. See Readme for installation instructions.')


def typer_func_parser[R](func: Callable[[str], R | None]) -> Callable[[str], R]:
    """Create a Typer argument parser from a function that returns a value or None."""

    @named_function('')
    def wrapper(arg: str) -> R:
        result = func(arg)
        if result is None:
            raise typer.BadParameter(arg)
        return result

    return wrapper


def typer_enum_options(enum: type[Enum]) -> str:
    """Return a string of the enum options for use in Typer help messages."""
    return ', '.join(f'"{member.lower()}"' for member in enum.__members__)


def typer_enum_parser(enum: type[Enum]) -> Callable[[str | Enum], Enum]:
    """Create a Typer argument parser from an Enum type."""

    @named_function('enum')
    def wrapper(arg: str | Enum) -> Enum:
        if isinstance(arg, Enum):  # Due to https://github.com/tiangolo/typer/discussions/720
            return arg
        try:
            return enum[arg.upper()]
        except KeyError:
            raise typer.BadParameter(
                f'{arg} (expected one of {typer_enum_options(enum)})',
            )

    return wrapper


@named_function('regex')
def typer_regex_parser(arg: str) -> re.Pattern:
    """Parse a regular expression string and return a compiled pattern or raise a Typer error."""
    try:
        return re.compile(arg)
    except re.error as e:
        raise typer.BadParameter(str(e))


def typer_enum_autocompletion(enum: type[Enum]) -> Callable[[str], list[str]]:
    """Create a Typer autocompletion function from an Enum type."""

    def wrapper(prefix: str) -> list[str]:
        lower = prefix.lower()
        return [
            replace_prefix_casing(member.lower(), prefix)
            for member in enum.__members__ if member.lower().startswith(lower)
        ]

    return wrapper


class TyperExtraOption:
    """A class for defining Typer options which are required in some cases but optional in others."""
    _prompt: str | None
    _data: str | None

    def __init__(self, *, prompt: str | None = None, data: str | None = None) -> None:
        self._prompt = prompt
        self._data = data

    @staticmethod
    @named_function('text')
    def parser(arg: str | TyperExtraOption) -> TyperExtraOption:
        # See https://github.com/tiangolo/typer/discussions/720
        return arg if isinstance(arg, TyperExtraOption) else TyperExtraOption(data=arg)

    @staticmethod
    def default_factory(prompt: str) -> Callable[[], TyperExtraOption]:
        return lambda: TyperExtraOption(prompt=prompt)

    def get(self) -> str:
        if self._data is None:
            self._data = typer.prompt(self._prompt)
        return self._data

    def promise(self) -> Promise[str]:
        return lambda: self.get()



from __future__ import annotations

from .table_str_enum import TableStrEnum

__all__ = [
    'Pos',
]


class Pos(TableStrEnum):
    NOUNS = 'nouns'
    VERBS = 'verbs'
    ADJECTIVES = 'adjectives'
    ADVERBS = 'adverbs'
    INTERJECTIONS = 'interjections'
    NUMERALS = 'numerals'
    PRONOUNS = 'pronouns'
    ADPOSITIONS = 'adpositions'
    CONJUNCTIONS = 'conjunctions'
    PARTICLES = 'particles'
    DETERMINERS = 'determiners'
    AUXILIARIES = 'auxiliaries'

    @staticmethod
    def from_upos(upos: str) -> Pos | None:
        match upos:
            case 'NOUN':
                return Pos.NOUNS
            case 'PROPN':
                return Pos.NOUNS
            case 'VERB':
                return Pos.VERBS
            case 'ADJ':
                return Pos.ADJECTIVES
            case 'ADV':
                return Pos.ADVERBS
            case 'INTJ':
                return Pos.INTERJECTIONS
            case 'NUM':
                return Pos.NUMERALS
            case 'PRON':
                return Pos.PRONOUNS
            case 'ADP':
                return Pos.ADPOSITIONS
            case 'CCONJ':
                return Pos.CONJUNCTIONS
            case 'SCONJ':
                return Pos.CONJUNCTIONS
            case 'PART':
                return Pos.PARTICLES
            case 'DET':
                return Pos.DETERMINERS
            case 'AUX':
                return Pos.AUXILIARIES
            case _:
                return None



from __future__ import annotations
from datetime import datetime, timedelta

from sqlalchemy import ForeignKey, Index
from sqlalchemy.orm import Mapped, mapped_column, relationship

from reling.db.base import Base
from reling.db.enums import Level
from .languages import Language

__all__ = [
    'Text',
    'TextExam',
    'TextExamResult',
    'TextSentence',
    'TextSentenceTranslation',
]


class Text(Base):
    __tablename__ = 'texts'

    id: Mapped[str] = mapped_column(primary_key=True)
    language_id: Mapped[str] = mapped_column(ForeignKey(Language.id))
    language: Mapped[Language] = relationship(Language)
    level: Mapped[Level]
    topic: Mapped[str]
    style: Mapped[str]
    created_at: Mapped[datetime]
    archived_at: Mapped[datetime | None]
    sentences: Mapped[list[TextSentence]] = relationship(
        'TextSentence',
        order_by='TextSentence.index',
        passive_deletes=True,
    )
    sentence_translations: Mapped[list[TextSentenceTranslation]] = relationship(
        'TextSentenceTranslation',
        passive_deletes=True,
    )
    exams: Mapped[list[TextExam]] = relationship('TextExam', passive_deletes=True)

    __table_args__ = (
        Index('text_chronological', 'archived_at', 'created_at'),
    )


class TextSentence(Base):
    __tablename__ = 'text_sentences'

    text_id: Mapped[str] = mapped_column(ForeignKey(Text.id, onupdate='CASCADE', ondelete='CASCADE'), primary_key=True)
    index: Mapped[int] = mapped_column(primary_key=True)
    sentence: Mapped[str]


class TextSentenceTranslation(Base):
    __tablename__ = 'text_sentence_translations'

    text_id: Mapped[str] = mapped_column(ForeignKey(Text.id, onupdate='CASCADE', ondelete='CASCADE'), primary_key=True)
    language_id: Mapped[str] = mapped_column(ForeignKey(Language.id), primary_key=True)
    text_sentence_index: Mapped[int] = mapped_column(primary_key=True)
    sentence: Mapped[str]


class TextExam(Base):
    __tablename__ = 'text_exams'

    id: Mapped[str] = mapped_column(primary_key=True)
    text_id: Mapped[str] = mapped_column(ForeignKey(Text.id, onupdate='CASCADE', ondelete='CASCADE'))
    text: Mapped[Text] = relationship(Text, viewonly=True)
    source_language_id: Mapped[str] = mapped_column(ForeignKey(Language.id))
    source_language: Mapped[Language] = relationship(Language, foreign_keys=source_language_id)
    target_language_id: Mapped[str] = mapped_column(ForeignKey(Language.id))
    target_language: Mapped[Language] = relationship(Language, foreign_keys=target_language_id)
    read_source: Mapped[bool]
    read_target: Mapped[bool]
    listened: Mapped[bool]
    started_at: Mapped[datetime]
    finished_at: Mapped[datetime]
    results: Mapped[list[TextExamResult]] = relationship(
        'TextExamResult',
        order_by='TextExamResult.text_sentence_index',
        passive_deletes=True,
    )

    __table_args__ = (
        Index('text_exam_source_language', 'source_language_id', 'started_at'),
        Index('text_exam_target_language', 'target_language_id', 'started_at'),
    )

    @property
    def item_id(self) -> str:
        return self.text_id

    @property
    def duration(self) -> timedelta:
        return self.finished_at - self.started_at


class TextExamResult(Base):
    __tablename__ = 'text_exam_results'

    text_exam_id: Mapped[str] = mapped_column(
        ForeignKey(TextExam.id, onupdate='CASCADE', ondelete='CASCADE'),
        primary_key=True,
    )
    text_sentence_index: Mapped[int] = mapped_column(primary_key=True)
    answer: Mapped[str]
    suggested_answer: Mapped[str | None]
    score: Mapped[int]



from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Callable, cast

from rich.text import Text

from reling.db import single_session
from reling.db.models import DialogueExam, Language, TextExam
from reling.utils.time import format_time_delta
from reling.utils.tables import build_table, GROUPING_COLUMN, print_table
from .checkpoints import colorize, since
from .filter import get_filter
from .modalities import Modality
from .progress import progress

__all__ = [
    'display_stats',
]

NA = 'N/A'

METRIC = 'Metric'
TEXTS = 'Texts'
DIALOGUES = 'Dialogues'
OVERALL = 'Overall'

TIME_SPENT = 'Total exam time'
EXAMS_TAKEN = 'Exams taken'
SENTENCES_EVALUATED = 'Sentences evaluated'
FIRST_TIME_EXAMS_TAKEN = 'First-time exams'
FIRST_TIME_SENTENCES_EVALUATED = 'Sents in first-time exams'
AVG_FTE_TIME_PER_SENTENCE = 'Avg. time/sent, first-time'
AVG_FTS_SCORE = 'Avg. score/sent, first-time'


@dataclass
class Stats:
    time_spent: timedelta = timedelta(0)
    exams_taken: int = 0
    sentences_evaluated: int = 0
    first_time_exams_taken: int = 0
    first_time_sentences_evaluated: int = 0
    fte_time_spent: timedelta = timedelta(0)
    fts_scores_total: float = 0.0

    def format_time_spent(self) -> str:
        return format_time_delta(self.time_spent)

    def format_exams_taken(self) -> str:
        return str(self.exams_taken)

    def format_sentences_evaluated(self) -> str:
        return str(self.sentences_evaluated)

    def format_first_time_exams_taken(self) -> str:
        return str(self.first_time_exams_taken)

    def format_first_time_sentences_evaluated(self) -> str:
        return str(self.first_time_sentences_evaluated)

    def format_avg_fte_time_per_sentence(self) -> str:
        return (format_time_delta(self.fte_time_spent / self.first_time_sentences_evaluated)
                if self.first_time_sentences_evaluated > 0 else NA)

    def format_avg_fts_score(self) -> str:
        return (f'{self.fts_scores_total / self.first_time_sentences_evaluated:.2f}'
                if self.first_time_sentences_evaluated > 0 else NA)


@dataclass
class TypeStats:
    text: Stats = field(default_factory=Stats)
    dialogue: Stats = field(default_factory=Stats)
    overall: Stats = field(default_factory=Stats)


@dataclass
class Checkpoint:
    start_time: datetime
    stats: TypeStats


@dataclass
class PeriodStats:
    all_time: TypeStats
    checkpoints: list[Checkpoint]


def get_relevant_periods(stats: PeriodStats, exam: TextExam | DialogueExam) -> list[TypeStats]:
    """Return the relevant statistics periods for the given exam."""
    return [
        stats.all_time,
        *(checkpoint.stats for checkpoint in stats.checkpoints if checkpoint.start_time <= exam.started_at),
    ]


def get_relevant_types(stats: TypeStats, exam: TextExam | DialogueExam) -> list[Stats]:
    """Return the relevant types for the given exam."""
    return [
        stats.overall,
        stats.text if isinstance(exam, TextExam) else stats.dialogue,
    ]


def update_single_stats(stats: Stats, exam: TextExam | DialogueExam, is_first_exam: bool) -> None:
    """Update the statistics with the given exam."""
    sentences_evaluated = len(cast(list, exam.results))
    scores_total = sum(result.score for result in exam.results)

    stats.time_spent += exam.duration
    stats.exams_taken += 1
    stats.sentences_evaluated += sentences_evaluated
    if is_first_exam:
        stats.first_time_exams_taken += 1
        stats.first_time_sentences_evaluated += sentences_evaluated
        stats.fte_time_spent += exam.duration
        stats.fts_scores_total += scores_total


def update_stats(stats: PeriodStats, exam: TextExam | DialogueExam, is_first_exam: bool) -> None:
    """Update the statistics with the given exam."""
    for period_stats in get_relevant_periods(stats, exam):
        for type_stats in get_relevant_types(period_stats, exam):
            update_single_stats(type_stats, exam, is_first_exam)


def compute_stats(language: Language, modality: Modality, checkpoints: list[datetime]) -> PeriodStats:
    """Compute regular statistics for the given language and modality."""
    with single_session() as session:
        stats = PeriodStats(
            all_time=TypeStats(),
            checkpoints=[Checkpoint(start_time=checkpoint, stats=TypeStats()) for checkpoint in checkpoints],
        )
        for model in [TextExam, DialogueExam]:
            seen_item_ids: set[str] = set()
            condition = get_filter(language, modality, model)
            for exam in progress(
                session.query(model).filter(condition).order_by(model.started_at),
                total=session.query(model).filter(condition).count(),
                modality=modality,
                model=model,
            ):
                update_stats(stats, exam, is_first_exam=exam.item_id not in seen_item_ids)
                seen_item_ids.add(exam.item_id)
        return stats


def build_stats_section(
        stats: PeriodStats,
        title: str,
        formatter: Callable[[Stats], str],
) -> list[dict[str, str | Text]]:
    """Build a section of statistics for the given title and formatter."""
    return [{
        GROUPING_COLUMN: title,
        METRIC: title,
        TEXTS: formatter(stats.all_time.text),
        DIALOGUES: formatter(stats.all_time.dialogue),
        OVERALL: formatter(stats.all_time.overall),
    }, *[{
        GROUPING_COLUMN: title,
        METRIC: colorize(since(checkpoint.start_time)),
        TEXTS: colorize(formatter(checkpoint.stats.text)),
        DIALOGUES: colorize(formatter(checkpoint.stats.dialogue)),
        OVERALL: colorize(formatter(checkpoint.stats.overall)),
    } for checkpoint in stats.checkpoints]]


def print_stats(stats: PeriodStats, modality: Modality) -> None:
    """Print the provided statistics."""
    print_table(build_table(
        title=modality.to_table_title(),
        headers=[
            METRIC,
            TEXTS,
            DIALOGUES,
            OVERALL,
        ],
        justify={
            METRIC: 'left',
            TEXTS: 'right',
            DIALOGUES: 'right',
            OVERALL: 'right',
        },
        data=[
            row
            for title, formatter in [
                (TIME_SPENT, Stats.format_time_spent),
                (EXAMS_TAKEN, Stats.format_exams_taken),
                (SENTENCES_EVALUATED, Stats.format_sentences_evaluated),
                (FIRST_TIME_EXAMS_TAKEN, Stats.format_first_time_exams_taken),
                (FIRST_TIME_SENTENCES_EVALUATED, Stats.format_first_time_sentences_evaluated),
                (AVG_FTE_TIME_PER_SENTENCE, Stats.format_avg_fte_time_per_sentence),
                (AVG_FTS_SCORE, Stats.format_avg_fts_score),
            ]
            for row in build_stats_section(stats, title, formatter)
        ],
        group_by=[GROUPING_COLUMN],
    ))


def display_stats(language: Language, modality: Modality, checkpoints: list[datetime]) -> None:
    """Display regular statistics for the given language and modality."""
    stats = compute_stats(language, modality, checkpoints)
    print_stats(stats, modality)



from __future__ import annotations
from dataclasses import dataclass
from typing import cast

from rich.console import Console
from rich.text import Text

from reling.tts import TTSVoiceClient
from reling.types import Reader, Speed
from reling.utils.console import clear_current_line
from reling.utils.prompts import enter_to_continue, Prompt, PromptOption
from reling.utils.values import coalesce, ensure_not_none
from .colors import fade

__all__ = [
    'output',
    'SentenceData',
]

NA = fade('N/A')

PROMPT_TITLE = 'Play'
NORMAL_SPEED = 'normal speed'
SLOWLY = 'slowly'
REPLAY = 'replay'


@dataclass
class SentenceData:
    print_text: str | Text | None = None
    print_prefix: str | Text | None = None
    reader: Reader | None = None
    reader_id: str | None = None

    @staticmethod
    def from_tts(
            text: str | None,
            client: TTSVoiceClient | None,
            *,
            print_text: str | Text | None = None,
            print_prefix: str | Text | None = None,
            reader_id: str | None = None,
    ) -> SentenceData:
        return SentenceData(
            print_text=coalesce(print_text, text),
            print_prefix=print_prefix,
            reader=client.get_reader(text) if client and (text is not None) else None,
            reader_id=reader_id,
        )


@dataclass
class ReaderWithSpeed:
    reader: Reader
    speed: Speed


@dataclass
class ReaderWithId:
    reader: Reader
    id: str


def add_single_sentence_options(prompt: Prompt[ReaderWithSpeed], reader: Reader) -> None:
    """Attach the options for a single sentence to the prompt: '[n]ormal speed | [s]lowly'."""
    prompt.add_option(PromptOption(
        description=NORMAL_SPEED,
        action=ReaderWithSpeed(reader, Speed.NORMAL),
    ))
    prompt.add_option(PromptOption(
        description=SLOWLY,
        action=ReaderWithSpeed(reader, Speed.SLOW),
    ))


def add_multi_sentence_options(prompt: Prompt[ReaderWithSpeed], readers: list[ReaderWithId]) -> None:
    """Attach the options for multiple sentences to the prompt: '[i]mproved | [is] | [o]riginal | [os]'."""
    for reader in readers:
        prompt.add_option(PromptOption(
            description=reader.id,
            action=ReaderWithSpeed(reader.reader, Speed.NORMAL),
            modifiers={
                SLOWLY: ReaderWithSpeed(reader.reader, Speed.SLOW),
            },
        ))


def construct_prompt(
        sentences_with_readers: list[SentenceData],
        current: ReaderWithSpeed | None,
        multi_sentence: bool,
) -> Prompt:
    """
    Construct a prompt for the user to choose the next sentence to read and the speed of the reading.
    :raises ValueError: If reader_id is not provided for a sentence with a reader in a multi-sentence output.
    """
    prompt = Prompt(PROMPT_TITLE)
    if multi_sentence:
        add_multi_sentence_options(prompt, [
            ReaderWithId(
                reader=ensure_not_none(sentence.reader),
                id=ensure_not_none(sentence.reader_id),
            )
            for sentence in sentences_with_readers
        ])
    else:
        add_single_sentence_options(prompt, sentences_with_readers[0].reader)
    if current:
        prompt.add_option(PromptOption(
            description=REPLAY,
            action=current,
        ))
    return prompt


def output(*sentences: SentenceData) -> None:
    """
    Output the sentences, reading them if a reader is provided.
    If multiple readers are provided, the user can choose which sentence to read next.
    The user can also choose the speed of the reading.

    :raises ValueError: If reader_id is not provided for a sentence with a reader in a multi-sentence output.
    """
    console = Console(highlight=False)
    for sentence in sentences:
        console.print(sentence.print_prefix or '', end='')
        console.print(coalesce(sentence.print_text, cast(str | Text, NA)))
    multi_sentence = len(sentences) > 1
    if sentences_with_readers := [sentence for sentence in sentences if sentence.reader]:
        current = ReaderWithSpeed(sentences_with_readers[0].reader, Speed.NORMAL) if len(sentences) == 1 else None
        while True:
            if current:
                try:
                    current.reader(current.speed)
                except KeyboardInterrupt:
                    pass
                clear_current_line()  # Otherwise the input made during the reading will get displayed twice
            current = construct_prompt(sentences_with_readers, current, multi_sentence).prompt()
            if not current:
                break
    elif multi_sentence:
        enter_to_continue()



from __future__ import annotations
from dataclasses import dataclass
from functools import lru_cache
from typing import cast, TYPE_CHECKING

if TYPE_CHECKING:
    from stanza import Pipeline

from reling.db import Session, single_session
from reling.db.models import GrammarCacheSentence, GrammarCacheWord, Language
from reling.utils.ids import generate_id
from .typer import typer_raise, typer_raise_import

__all__ = [
    'Analyzer',
    'WordInfo',
]

PROCESSORS = 'tokenize,pos,lemma'

EXCLUDED_UPOS = {'PUNCT', 'SYM', 'X'}


@dataclass
class WordInfo:
    text: str
    lemma: str
    upos: str


@dataclass
class Analyzer:
    _nlp: Pipeline
    language: Language

    @staticmethod
    @lru_cache
    def get(language: Language) -> Analyzer:
        """Get a Stanza pipeline for the specified language."""
        try:
            import stanza
            from stanza.resources.common import UnknownLanguageError
        except ImportError:
            raise typer_raise_import('Stanza')
        import logging
        stanza_logger = logging.getLogger('stanza')
        stanza_logger.setLevel(logging.ERROR)
        try:
            stanza.download(language.short_code)
        except UnknownLanguageError:
            raise typer_raise(f'{language.name} is not supported by Stanza.')
        return Analyzer(
            _nlp=stanza.Pipeline(language.short_code, processors=PROCESSORS),
            language=language,
        )

    def _get_analysis_from_cache(self, session: Session, sentence: str) -> list[WordInfo] | None:
        """Get the analysis of a sentence from the cache."""
        return [
            WordInfo(
                text=cast(str, word.text),
                lemma=cast(str, word.lemma),
                upos=cast(str, word.upos),
            )
            for word in (session.query(GrammarCacheWord)
                         .filter_by(sentence_id=cached_sentence.id)
                         .order_by(GrammarCacheWord.index).all())
        ] if (cached_sentence := session.query(GrammarCacheSentence).filter_by(
            language_id=self.language.id,
            sentence=sentence,
        ).first()) else None

    def _put_analysis_to_cache(self, session: Session, sentence: str, words: list[WordInfo]) -> None:
        """Put the analysis of a sentence to the cache."""
        sentence_id = generate_id()
        cached_sentence = GrammarCacheSentence(
            id=sentence_id,
            language_id=self.language.id,
            sentence=sentence,
        )
        session.add(cached_sentence)
        for index, word in enumerate(words):
            session.add(GrammarCacheWord(
                sentence_id=sentence_id,
                index=index,
                text=word.text,
                lemma=word.lemma,
                upos=word.upos,
            ))
        session.commit()

    def _do_analyze(self, sentence: str) -> list[WordInfo]:
        """Analyze a sentence."""
        for section in self._nlp(sentence).sentences:
            return [
                WordInfo(
                    text=word.text,
                    lemma=word.lemma,
                    upos=word.upos,
                )
                for word in section.words
                if word.upos not in EXCLUDED_UPOS
            ]

    def analyze(self, sentence: str) -> list[WordInfo]:
        """Analyze a sentence and cache the result."""
        with single_session() as session:
            if (words := self._get_analysis_from_cache(session, sentence)) is not None:
                return words
            else:
                words = self._do_analyze(sentence)
                self._put_analysis_to_cache(session, sentence, words)
                return words



from datetime import datetime

from rich.text import Text

from reling.helpers.colors import fade
from reling.utils.time import format_time

__all__ = [
    'colorize',
    'since',
]

SINCE_TEXT = 'since {checkpoint}'
TODAY_TEXT = 'today'


def since(checkpoint: datetime) -> str:
    """Return a string indicating the time period since the given checkpoint."""
    return SINCE_TEXT.format(checkpoint=format_time(checkpoint, omit_zero_time=True))


def colorize(text: str) -> Text:
    """Colorize the given checkpoint text for display."""
    return fade(text)



from dataclasses import dataclass
from typing import Self

__all__ = [
    'WordWithSense',
]


@dataclass
class WordWithSense:
    word: str
    sense: str | None

    DELIMITER_WITH_WHITE_SPACE = ': '

    @classmethod
    def parse(cls, text: str) -> Self:
        """
        Parse the word and sense from the given string:
        "word: sense" -> WordWithSense("word", "sense")
        "word" -> WordWithSense("word", None)
        """
        delimiter = cls.DELIMITER_WITH_WHITE_SPACE.strip()
        if delimiter in text:
            word, sense = text.split(delimiter, 1)
            return cls(word.strip(), sense.strip())
        else:
            return cls(text.strip(), None)

    def format(self) -> str:
        return f'"{self.word}"' + (f' ({self.sense})' if self.sense else '')



from pathlib import Path
import shelve

__all__ = [
    'get_value',
    'init_shelf',
    'set_value',
]

FILENAME: Path | None = None


def init_shelf(filename: Path) -> None:
    global FILENAME
    FILENAME = filename


def get_filename() -> str:
    if FILENAME is None:
        raise RuntimeError('Shelf is not initialized')
    return str(FILENAME.absolute())


def get_value(key: str) -> str | None:
    with shelve.open(get_filename()) as db:
        return db.get(key)


def set_value(key: str, value: str | None) -> None:
    with shelve.open(get_filename()) as db:
        db[key] = value



from itertools import starmap
from typing import Generator

from reling.db.enums import ContentCategory, Gender, Level
from reling.db.models import Language
from reling.gpt import GPTClient
from reling.types import DialogueExchangeData, WordWithSense
from reling.utils.english import pluralize
from reling.utils.iterables import pair_items
from reling.utils.transformers import omit_empty, remove_numbering, slugify, strip

__all__ = [
    'generate_dialogue_exchanges',
    'generate_id',
    'generate_text_sentences',
]

# We ask the model to number each sentence/line of its response because this approach makes it more reliable in placing
# sentences on new lines and ensures that there are exactly the specified number of sentences in the response.


def build_level_prompt(level: Level, category: ContentCategory) -> str:
    """Return a prompt section describing the level of the content."""

    def get_sentences_description() -> str:
        match level:
            case Level.BASIC:
                return 'very simple'
            case Level.INTERMEDIATE:
                return 'rather simple'
            case Level.ADVANCED:
                return 'complex'

    def get_vocabulary_description() -> str:
        match level:
            case Level.BASIC:
                return 'very basic'
            case Level.INTERMEDIATE:
                return 'intermediate'
            case Level.ADVANCED:
                return 'advanced'

    return (f'The {category.value} should consist of {get_sentences_description()} sentences '
            f'and use {get_vocabulary_description()} vocabulary.')


def build_include_prompt(include: list[WordWithSense]) -> list[str]:
    """Return a prompt section describing the words or phrases to include in the content."""
    return [] if len(include) == 0 else [
        f'Include the following {pluralize('word', len(include))} or {pluralize('phrase', len(include))}:',
        *(item.format() for item in include),
    ]


def generate_text_sentences(
        gpt: GPTClient,
        num_sentences: int,
        language: Language,
        level: Level,
        topic: str,
        style: str,
        include: list[WordWithSense],
) -> Generator[str, None, None]:
    return gpt.ask(
        '\n'.join([
            f'Generate a text in {language.name} consisting of {num_sentences} {pluralize('sentence', num_sentences)}.',
            f'The text should be about {topic} and be written in the style of {style}.',
            f'Do not include any additional text; only generate the text as specified.',
            f'Number each sentence and put each sentence on a new line.',
            build_level_prompt(level, ContentCategory.TEXT),
            *build_include_prompt(include),
        ]),
        transformers=[strip, omit_empty, remove_numbering],
    )


def generate_dialogue_exchanges(
        gpt: GPTClient,
        num_exchanges: int,
        language: Language,
        level: Level,
        user_gender: Gender,
        speaker: str,
        speaker_gender: Gender,
        topic: str | None,
        include: list[WordWithSense],
) -> Generator[DialogueExchangeData, None, None]:
    return iter(starmap(DialogueExchangeData, pair_items(gpt.ask(
        '\n'.join([
            f'Generate a dialogue in {language.name} consisting of {num_exchanges * 2} sentences.',
            f'The dialogue should be between two speakers, {speaker} and me.',
            *([f'The dialogue should be about {topic}.'] if topic else []),
            f'Do not include any additional text; only generate the text as specified.',
            f'Number each sentence and put each sentence on a new line.',
            f'The first, third, etc. sentences should be spoken by {speaker} ({speaker_gender.describe()}).'
            f'The second, fourth, etc. sentences should be spoken by me ({user_gender.describe()}).'
            f'Do not prefix the sentences with the speakers\' names.',
            build_level_prompt(level, ContentCategory.DIALOGUE),
            *build_include_prompt(include),
        ]),
        transformers=[strip, omit_empty, remove_numbering],
    ))))


def generate_id(
        gpt: GPTClient,
        sentences: list[str],
) -> str:
    return (list(gpt.ask(
        '\n'.join([
            'What should the following text be called in English?',
            '"""',
            *sentences,
            '"""',
            'The name should be a short, descriptive title.',
            'Do not include any additional text; only generate the English name as specified.',
        ]),
        transformers=[slugify],
    )) or [''])[0]



from reling.app.app import app
from reling.app.types import CHECKPOINT_OPT, COMPREHENSION_OPT, GRAMMAR_OPT, LANGUAGE_ARG, PRODUCTION_OPT
from reling.utils.time import local_to_utc
from .grammar_stats import display_stats as display_grammar_stats
from .modalities import Modality
from .regular_stats import display_stats as display_regular_stats

__all__ = [
    'stats',
]


@app.command()
def stats(
        language: LANGUAGE_ARG,
        grammar: GRAMMAR_OPT = False,
        comprehension: COMPREHENSION_OPT = False,
        production: PRODUCTION_OPT = False,
        checkpoint: CHECKPOINT_OPT = None,
) -> None:
    """Show learning statistics for a specific language."""
    comprehension, production = comprehension or not production, production or not comprehension
    display_stats = display_grammar_stats if grammar else display_regular_stats
    checkpoint_dates = list(map(local_to_utc, checkpoint or []))
    for (modality, should_display) in [
        (Modality.COMPREHENSION, comprehension),
        (Modality.PRODUCTION, production),
    ]:
        if should_display:
            print()
            display_stats(language, modality, checkpoint_dates)



from itertools import islice, starmap
from typing import cast, Generator

from reling.app.config import MAX_SCORE
from reling.app.exceptions import AlgorithmException
from reling.db.enums import ContentCategory
from reling.db.models import Language
from reling.gpt import GPTClient
from reling.helpers.scoring import calculate_diff_score
from reling.types import DialogueExchangeData, Promise
from reling.utils.english import pluralize
from reling.utils.iterables import group_items
from reling.utils.transformers import add_numbering, apply, omit_empty, remove_numbering, strip
from .types import ExchangeWithTranslation, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'score_dialogue_translations',
    'score_text_translations',
]

NA = 'N/A'
EMPTY_TRANSLATION = '<empty>'


def build_prompt(
        category: ContentCategory,
        source_language: Language,
        target_language: Language,
        blocks: list[str],
        translations: list[str],
) -> str:
    """Build a prompt for scoring translations."""
    # Speaker turns in dialogues are "graded" as well so that the model appreciates the context.
    n = len(blocks)
    return '\n'.join([
        f'Below {'is' if n == 1 else 'are'} {n} {pluralize('sentence', n)} from a {category.value} '
        f'in {source_language.name} along with {'its' if n == 1 else 'their'} {pluralize('translation', n)} '
        f'into {target_language.name} made by a language learner.',

        f'Score {'the' if n == 1 else 'each'} translation on a scale from 0 to {MAX_SCORE}. '
        f'If {'the' if n == 1 else 'a'} translation is empty, very short, or poor, assign a low score. ',
        f'If the translation is less than perfect, suggest a minimally modified version that would '
        f'deserve a {MAX_SCORE}.',

        f'{'Provide' if n == 1 else 'For each translation, provide'} your feedback on exactly four lines:',
        f'- original sentence on the first line;',  # The first two lines help improve the model's performance
        f'- learner\'s translation on the second line;',
        f'- score (just the number) on the third line;',
        f'- suggested modified translation (or "{NA}") on the fourth line.',

        *([f'Provide this feedback for each of the {n} translations.'] if n > 1 else []),
        f'Say nothing else.',
        f'',
        f'The original {category.value} is:',
        *apply(add_numbering, blocks),
        f'',
        f'The translations are:',
        *apply(add_numbering, [translation.strip() or EMPTY_TRANSLATION for translation in translations]),
    ])


def ask_and_parse(gpt: Promise[GPTClient], prompt: str) -> Generator[ScoreWithSuggestion, None, None]:
    """
    Ask the model to score translations and parse the output.
    :raises AlgorithmException: If there is an issue with the output of the model.
    """
    for _, _, string_score, suggestion in group_items(gpt().ask(
        prompt,
        creative=False,
        transformers=[strip, omit_empty, remove_numbering],
    ), 4):
        try:
            score = int(string_score)
        except ValueError:
            raise AlgorithmException(f'Could not parse the score as an integer from the model output: {string_score}.')
        if score < 0 or score > MAX_SCORE:
            raise AlgorithmException(f'The score {score} given by the model is not in the range from 0 to {MAX_SCORE}.')
        yield ScoreWithSuggestion(
            score=score,
            suggestion=(stripped or None) if (stripped := suggestion.strip()) != NA else None,
        )


def compare_strings(
        provided_translation: str,
        original_translation: str,
        score: ScoreWithSuggestion,
) -> ScoreWithSuggestion:
    """
    Return the highest score among the original score and the scores calculated using the diffs between
    the provided translation and both the original and suggested translations.
    """
    return ScoreWithSuggestion(
        score=max([score.score] + [calculate_diff_score(provided_translation, corrected) for corrected in filter(None, [
            original_translation,
            score.suggestion,
        ])]),
        suggestion=score.suggestion,
    )


def score_text_translations(
        gpt: Promise[GPTClient],
        sentences: list[SentenceWithTranslation],
        original_translations: list[str],
        source_language: Language,
        target_language: Language,
        offline: bool,
) -> Generator[ScoreWithSuggestion, None, None]:
    """
    Score the translations of a text and provide suggestions for improvement.
    :raises AlgorithmException: If there is an issue with the scoring algorithm.
    """
    if offline:
        for sentence, original_translation in zip(sentences, original_translations):
            yield ScoreWithSuggestion(
                score=calculate_diff_score(sentence.translation.text, original_translation),
                suggestion=original_translation,
            )
    else:
        prompt = build_prompt(
            category=ContentCategory.TEXT,
            source_language=source_language,
            target_language=target_language,
            blocks=[cast(str, sentence.sentence) for sentence in sentences],
            translations=[sentence.translation.text for sentence in sentences],
        )
        yield from starmap(compare_strings, zip(
            (sentence.translation.text for sentence in sentences),
            original_translations,
            ask_and_parse(gpt, prompt),
        ))


def score_dialogue_translations(
        gpt: Promise[GPTClient],
        exchanges: list[ExchangeWithTranslation],
        original_translations: list[DialogueExchangeData],
        source_language: Language,
        target_language: Language,
        offline: bool,
) -> Generator[ScoreWithSuggestion, None, None]:
    """
    Score the translations of user turns in a dialogue and provide suggestions for improvement.
    :raises AlgorithmException: If there is an issue with the scoring algorithm.
    """
    if offline:
        for exchange, original_translation in zip(exchanges, original_translations):
            yield ScoreWithSuggestion(
                score=calculate_diff_score(exchange.user_translation.text, original_translation.user),
                suggestion=original_translation.user,
            )
    else:
        prompt = build_prompt(
            category=ContentCategory.DIALOGUE,
            source_language=source_language,
            target_language=target_language,
            blocks=[turn for exchange in exchanges for turn in exchange.exchange.all()],
            translations=[turn
                          for exchange, original_translation in zip(exchanges, original_translations)
                          for turn in [original_translation.speaker, exchange.user_translation.text]],
        )
        yield from starmap(compare_strings, zip(
            (exchange.user_translation.text for exchange in exchanges),
            (original_translation.user for original_translation in original_translations),
            islice(ask_and_parse(gpt, prompt), 1, None, 2),
        ))



from reling.db.models import Language
from reling.types import Promise
from .gtts import GTTSClient
from .openai import OpenAITTSClient
from .tts_client import TTSClient, TTSVoiceClient
from .voices import Voice

__all__ = [
    'get_tts_client',
    'TTSClient',
    'TTSVoiceClient',
    'Voice',
]

GTTS_MODEL = '_gtts'


def get_tts_client(model: str, api_key: Promise[str], language: Language) -> TTSClient:
    if model == GTTS_MODEL:
        return GTTSClient(language)
    else:
        return OpenAITTSClient(api_key=api_key(), model=model)



from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path
from tempfile import mkstemp

from reling.types import Input, Transcriber
from reling.utils.console import clear_previous, input_and_erase, print_and_erase
from reling.utils.prompts import ENTER, format_shortcut, Prompt
from .wave import FILE_EXTENSION, play, record

__all__ = [
    'get_input',
    'TranscriberParams',
]


class Action(StrEnum):
    RE_RECORD = 're-record'
    LISTEN = 'listen'
    MANUAL_INPUT = 'manual input'


ENTER_TO_START_RECORDING = f'({ENTER} to start recording)'
RECORDING_REDO_SHORTCUT = Action.RE_RECORD.value[0]
RECORDING_UNTIL_ENTER = f'(recording... {ENTER} to stop, {format_shortcut(RECORDING_REDO_SHORTCUT)} to redo)'
TRANSCRIBING = '(transcribing...)'


@dataclass
class TranscriberParams:
    transcribe: Transcriber
    storage: Path


def get_manual_input(prompt: str) -> Input:
    return Input(input(prompt))


def get_temp_file(storage: Path) -> Path:
    return Path(mkstemp(dir=storage, suffix=FILE_EXTENSION)[1])


def do_record(prompt: str, file: Path) -> bool:
    """Record audio and return whether the user is satisfied with the recording."""
    with record(file):
        return input_and_erase(prompt + RECORDING_UNTIL_ENTER).strip().lower() != RECORDING_REDO_SHORTCUT.lower()


def do_transcribe(prompt: str, transcribe: Transcriber, file: Path) -> str:
    with print_and_erase(prompt + TRANSCRIBING):
        try:
            return transcribe(file)
        except KeyboardInterrupt:
            return ''


def get_audio_input(prompt: str, params: TranscriberParams) -> Input:
    """Get input from the user via audio recording, with optional re-recording, listening, and manual input."""
    input_and_erase(prompt + ENTER_TO_START_RECORDING)
    file = get_temp_file(params.storage)
    while True:
        if not do_record(prompt, file):
            continue
        transcription = do_transcribe(prompt, params.transcribe, file)
        print(prompt + transcription)
        while True:
            match Prompt.from_enum(Action).prompt():
                case Action.RE_RECORD:
                    clear_previous()
                    break
                case Action.LISTEN:
                    try:
                        play(file)
                    except KeyboardInterrupt:
                        pass
                case Action.MANUAL_INPUT:
                    clear_previous()
                    return get_manual_input(prompt)
                case None:
                    return Input(transcription, audio=file)
                case _:
                    assert False


def get_input(prompt: str = '', transcriber_params: TranscriberParams | None = None) -> Input:
    """Get input from the user, optionally via audio recording."""
    return get_audio_input(prompt, transcriber_params) if transcriber_params else get_manual_input(prompt)



__all__ = [
    'coalesce',
    'ensure_not_none',
]


def ensure_not_none[T](value: T | None) -> T:
    """Raise a ValueError if the value is None."""
    if value is None:
        raise ValueError('Value cannot be None.')
    return value


def coalesce[T](*values: T | None) -> T | None:
    """Return the first non-null value, if any."""
    return next((value for value in values if value is not None), None)



from __future__ import annotations
from datetime import datetime, timedelta

from sqlalchemy import ForeignKey, Index
from sqlalchemy.orm import Mapped, mapped_column, relationship

from reling.db.base import Base
from reling.db.enums import Gender, Level
from .languages import Language

__all__ = [
    'Dialogue',
    'DialogueExam',
    'DialogueExamResult',
    'DialogueExchange',
    'DialogueExchangeTranslation',
]


class Dialogue(Base):
    __tablename__ = 'dialogues'

    id: Mapped[str] = mapped_column(primary_key=True)
    language_id: Mapped[str] = mapped_column(ForeignKey(Language.id))
    language: Mapped[Language] = relationship(Language)
    level: Mapped[Level]
    speaker: Mapped[str]
    topic: Mapped[str | None]
    speaker_gender: Mapped[Gender]
    user_gender: Mapped[Gender]
    created_at: Mapped[datetime]
    archived_at: Mapped[datetime | None]
    exchanges: Mapped[list[DialogueExchange]] = relationship(
        'DialogueExchange',
        order_by='DialogueExchange.index',
        passive_deletes=True,
    )
    exchange_translations: Mapped[list[DialogueExchangeTranslation]] = relationship(
        'DialogueExchangeTranslation',
        passive_deletes=True,
    )
    exams: Mapped[list[DialogueExam]] = relationship('DialogueExam', passive_deletes=True)

    __table_args__ = (
        Index('dialogue_chronological', 'archived_at', 'created_at'),
    )


class DialogueExchange(Base):
    __tablename__ = 'dialogue_exchanges'

    dialogue_id: Mapped[str] = mapped_column(
        ForeignKey(Dialogue.id, onupdate='CASCADE', ondelete='CASCADE'),
        primary_key=True,
    )
    index: Mapped[int] = mapped_column(primary_key=True)
    speaker: Mapped[str]
    user: Mapped[str]


class DialogueExchangeTranslation(Base):
    __tablename__ = 'dialogue_exchange_translations'

    dialogue_id: Mapped[str] = mapped_column(
        ForeignKey(Dialogue.id, onupdate='CASCADE', ondelete='CASCADE'),
        primary_key=True,
    )
    language_id: Mapped[str] = mapped_column(ForeignKey(Language.id), primary_key=True)
    dialogue_exchange_index: Mapped[int] = mapped_column(primary_key=True)
    speaker: Mapped[str]
    user: Mapped[str]


class DialogueExam(Base):
    __tablename__ = 'dialogue_exams'

    id: Mapped[str] = mapped_column(primary_key=True)
    dialogue_id: Mapped[str] = mapped_column(ForeignKey(Dialogue.id, onupdate='CASCADE', ondelete='CASCADE'))
    dialogue: Mapped[Dialogue] = relationship(Dialogue, viewonly=True)
    source_language_id: Mapped[str] = mapped_column(ForeignKey(Language.id))
    source_language: Mapped[Language] = relationship(Language, foreign_keys=source_language_id)
    target_language_id: Mapped[str] = mapped_column(ForeignKey(Language.id))
    target_language: Mapped[Language] = relationship(Language, foreign_keys=target_language_id)
    read_source: Mapped[bool]
    read_target: Mapped[bool]
    listened: Mapped[bool]
    started_at: Mapped[datetime]
    finished_at: Mapped[datetime]
    results: Mapped[list[DialogueExamResult]] = relationship(
        'DialogueExamResult',
        order_by='DialogueExamResult.dialogue_exchange_index',
        passive_deletes=True,
    )

    __table_args__ = (
        Index('dialogue_exam_source_language', 'source_language_id', 'started_at'),
        Index('dialogue_exam_target_language', 'target_language_id', 'started_at'),
    )

    @property
    def item_id(self) -> str:
        return self.dialogue_id

    @property
    def duration(self) -> timedelta:
        return self.finished_at - self.started_at


class DialogueExamResult(Base):
    __tablename__ = 'dialogue_exam_results'

    dialogue_exam_id: Mapped[str] = mapped_column(
        ForeignKey(DialogueExam.id, onupdate='CASCADE', ondelete='CASCADE'),
        primary_key=True,
    )
    dialogue_exchange_index: Mapped[int] = mapped_column(primary_key=True)
    answer: Mapped[str]
    suggested_answer: Mapped[str | None]
    score: Mapped[int]



from pathlib import Path

__all__ = [
    'LANGUAGES_PATH',
    'SPEAKERS_PATH',
    'STYLES_PATH',
    'TOPICS_PATH',
]

DATA = Path(__file__).parent / 'data'

LANGUAGES_PATH = DATA / 'languages.csv'
SPEAKERS_PATH = DATA / 'speakers.csv'
STYLES_PATH = DATA / 'styles.csv'
TOPICS_PATH = DATA / 'topics.csv'



from sqlalchemy.orm import Mapped, mapped_column

from reling.db.base import Base
from reling.db.enums import ContentCategory

__all__ = [
    'IdIndex',
]


class IdIndex(Base):
    __tablename__ = 'id_index'

    id: Mapped[str] = mapped_column(primary_key=True)
    category: Mapped[ContentCategory]



from datetime import datetime

from reling.db import single_session
from reling.db.models import Dialogue, DialogueExam, DialogueExamResult, Language, Text, TextExam, TextExamResult
from reling.utils.ids import generate_id
from .types import ExchangeWithTranslation, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'save_dialogue_exam',
    'save_text_exam',
]


def save_text_exam(
        text: Text,
        source_language: Language,
        target_language: Language,
        read_source: bool,
        read_target: bool,
        listened: bool,
        started_at: datetime,
        finished_at: datetime,
        sentences: list[SentenceWithTranslation],
        results: list[ScoreWithSuggestion],
) -> None:
    """Save the results of a text exam."""
    with single_session() as session:
        exam = TextExam(
            id=generate_id(),
            text_id=text.id,
            source_language_id=source_language.id,
            target_language_id=target_language.id,
            read_source=read_source,
            read_target=read_target,
            listened=listened,
            started_at=started_at,
            finished_at=finished_at,
        )
        session.add(exam)
        for index, (sentence, result) in enumerate(zip(sentences, results)):
            session.add(TextExamResult(
                text_exam_id=exam.id,
                text_sentence_index=index,
                answer=sentence.translation.text,
                suggested_answer=result.suggestion,
                score=result.score,
            ))
        session.commit()


def save_dialogue_exam(
        dialogue: Dialogue,
        source_language: Language,
        target_language: Language,
        read_source: bool,
        read_target: bool,
        listened: bool,
        started_at: datetime,
        finished_at: datetime,
        exchanges: list[ExchangeWithTranslation],
        results: list[ScoreWithSuggestion],
) -> None:
    """Save the results of a dialogue exam."""
    with single_session() as session:
        exam = DialogueExam(
            id=generate_id(),
            dialogue_id=dialogue.id,
            source_language_id=source_language.id,
            target_language_id=target_language.id,
            read_source=read_source,
            read_target=read_target,
            listened=listened,
            started_at=started_at,
            finished_at=finished_at,
        )
        session.add(exam)
        for index, (exchange, result) in enumerate(zip(exchanges, results)):
            session.add(DialogueExamResult(
                dialogue_exam_id=exam.id,
                dialogue_exchange_index=index,
                answer=exchange.user_translation.text,
                suggested_answer=result.suggestion,
                score=result.score,
            ))
        session.commit()



from datetime import datetime
import re
from typing import Annotated

import typer

from reling.db.enums import ContentCategory, Gender, Level
from reling.db.helpers.content import find_content
from reling.db.helpers.ids import find_ids_by_prefix
from reling.db.helpers.languages import find_language, find_languages_by_prefix
from reling.db.models import Dialogue, Language, Text
from reling.helpers.typer import (
    TyperExtraOption,
    typer_enum_autocompletion,
    typer_enum_options,
    typer_enum_parser,
    typer_func_parser,
    typer_regex_parser,
)
from reling.types import WordWithSense
from reling.utils.time import DATE_FORMAT, TIME_FORMAT

__all__ = [
    'API_KEY',
    'ARCHIVE_OPT',
    'ASR_MODEL',
    'CHECKPOINT_OPT',
    'COMPREHENSION_OPT',
    'CONTENT_ARG',
    'CONTENT_CATEGORY_OPT',
    'FORCE_OPT',
    'GRAMMAR_OPT',
    'HIDE_PROMPTS_OPT',
    'IDS_ONLY_OPT',
    'INCLUDE_OPT',
    'LANGUAGE_ARG',
    'LANGUAGE_OPT',
    'LANGUAGE_OPT_ARG',
    'LANGUAGE_OPT_FROM',
    'LEVEL_OPT',
    'LISTEN_OPT',
    'MODEL',
    'NEW_ID_ARG',
    'OFFLINE_SCORING_OPT',
    'PRODUCTION_OPT',
    'READ_LANGUAGE_OPT',
    'READ_OPT',
    'REGEX_CONTENT_OPT',
    'SIZE_DIALOGUE_OPT',
    'SIZE_TEXT_OPT',
    'SPEAKER_GENDER_OPT',
    'SPEAKER_OPT',
    'STYLE_OPT',
    'TOPIC_OPT',
    'TTS_MODEL',
    'USER_GENDER',
]

ENV_PREFIX = 'RELING_'

LAST_CONTENT_MARKER = '.'

API_KEY = Annotated[TyperExtraOption, typer.Option(
    envvar=f'{ENV_PREFIX}API_KEY',
    parser=TyperExtraOption.parser,
    help='your OpenAI API key',
    default_factory=TyperExtraOption.default_factory(
        prompt='Enter your OpenAI API key',
    ),
)]

MODEL = Annotated[TyperExtraOption, typer.Option(
    envvar=f'{ENV_PREFIX}MODEL',
    parser=TyperExtraOption.parser,
    help='identifier for the GPT model to be used',
    default_factory=TyperExtraOption.default_factory(
        prompt='Enter the GPT model identifier',
    ),
)]

TTS_MODEL = Annotated[TyperExtraOption, typer.Option(
    envvar=f'{ENV_PREFIX}TTS_MODEL',
    parser=TyperExtraOption.parser,
    help='identifier for the TTS model to be used',
    default_factory=TyperExtraOption.default_factory(
        prompt='Enter the TTS model identifier',
    ),
)]

ASR_MODEL = Annotated[TyperExtraOption, typer.Option(
    envvar=f'{ENV_PREFIX}ASR_MODEL',
    parser=TyperExtraOption.parser,
    help='identifier for the ASR model to be used',
    default_factory=TyperExtraOption.default_factory(
        prompt='Enter the ASR model identifier',
    ),
)]

USER_GENDER = Annotated[Gender, typer.Option(
    envvar=f'{ENV_PREFIX}USER_GENDER',
    parser=typer_enum_parser(Gender),
    help=f'user\'s gender, one of: {typer_enum_options(Gender)} (to customize the generated content)',
    prompt=f'Enter your gender ({typer_enum_options(Gender)})',
    autocompletion=typer_enum_autocompletion(Gender),
)]

type TextOrDialogue = Text | Dialogue  # Typer does not yet support union types (except for Optional)

CONTENT_ARG = Annotated[TextOrDialogue, typer.Argument(
    parser=typer_func_parser(lambda content_id: find_content(content_id, LAST_CONTENT_MARKER)),
    help=f'ID of the text or dialogue ("{LAST_CONTENT_MARKER}" for the last used content)',
    autocompletion=find_ids_by_prefix,
)]

LANGUAGE_ARG = Annotated[Language, typer.Argument(
    parser=typer_func_parser(find_language),
    help='language code or name',
    autocompletion=find_languages_by_prefix,
)]

LANGUAGE_OPT_ARG = Annotated[Language | None, typer.Argument(
    parser=typer_func_parser(find_language),
    help='language code or name',
    autocompletion=find_languages_by_prefix,
)]

LANGUAGE_OPT = Annotated[Language | None, typer.Option(
    parser=typer_func_parser(find_language),
    help='language code or name',
    autocompletion=find_languages_by_prefix,
)]

LANGUAGE_OPT_FROM = Annotated[Language | None, typer.Option(
    '--from',  # `from` is a reserved keyword
    parser=typer_func_parser(find_language),
    help='language code or name',
    autocompletion=find_languages_by_prefix,
)]

CONTENT_CATEGORY_OPT = Annotated[ContentCategory | None, typer.Option(
    parser=typer_enum_parser(ContentCategory),
    help=f'content category, one of: {typer_enum_options(ContentCategory)}',
    autocompletion=typer_enum_autocompletion(ContentCategory),
)]

LEVEL_OPT = Annotated[Level | None, typer.Option(
    parser=typer_enum_parser(Level),
    help=f'level, one of: {typer_enum_options(Level)}',
    autocompletion=typer_enum_autocompletion(Level),
)]

TOPIC_OPT = Annotated[str | None, typer.Option(
    help='topic of the content, e.g., "food" or "zoology"',
)]

STYLE_OPT = Annotated[str | None, typer.Option(
    help='style of the text, e.g., "mystery" or "news article"',
)]

SPEAKER_OPT = Annotated[str | None, typer.Option(
    help='interlocutor in the dialogue, e.g., "waiter" or "friend"',
)]

SPEAKER_GENDER_OPT = Annotated[Gender | None, typer.Option(
    parser=typer_enum_parser(Gender),
    help=f'interlocutor\'s gender, one of: {typer_enum_options(Gender)}',
    autocompletion=typer_enum_autocompletion(Gender),
)]

SIZE_TEXT_OPT = Annotated[int, typer.Option(
    min=1,
    help='number of sentences in the text',
)]

SIZE_DIALOGUE_OPT = Annotated[int, typer.Option(
    min=1,
    help='number of exchanges in the dialogue',
)]

INCLUDE_OPT = Annotated[list[str] | None, typer.Option(
    help=f'word(s) or phrase(s) to be included in the content '
         f'(e.g., "bank" or "bank{WordWithSense.DELIMITER_WITH_WHITE_SPACE}financial institution" for disambiguation)',
)]

NEW_ID_ARG = Annotated[str, typer.Argument(
    help='new ID for the text or dialogue',
)]

REGEX_CONTENT_OPT = Annotated[re.Pattern | None, typer.Option(
    parser=typer_regex_parser,
    help='regular expression to filter results by ID, content, topic, style, or speaker',
)]

CHECKPOINT_OPT = Annotated[list[datetime] | None, typer.Option(
    help='starting date(s) or time(s) to add statistics checkpoints',
    formats=[DATE_FORMAT, TIME_FORMAT],
)]

READ_LANGUAGE_OPT = Annotated[list[Language] | None, typer.Option(
    parser=typer_func_parser(find_language),
    help='language(s) to read the content out loud in',
    autocompletion=find_languages_by_prefix,
)]

READ_OPT = Annotated[bool | None, typer.Option(
    help='Read the content out loud.',
)]

LISTEN_OPT = Annotated[bool | None, typer.Option(
    help='Record the response as audio and transcribe it into text.',
)]

HIDE_PROMPTS_OPT = Annotated[bool | None, typer.Option(
    help='Hide the original language text and interlocutor\'s turn.',
)]

OFFLINE_SCORING_OPT = Annotated[bool | None, typer.Option(
    help='Score answers using an offline algorithm.',
)]

GRAMMAR_OPT = Annotated[bool | None, typer.Option(
    help='Include statistics on learned words.',
)]

COMPREHENSION_OPT = Annotated[bool | None, typer.Option(
    help='Compute only comprehension-related statistics.',
)]

PRODUCTION_OPT = Annotated[bool | None, typer.Option(
    help='Compute only production-related statistics.',
)]

ARCHIVE_OPT = Annotated[bool | None, typer.Option(
    help='Search within archived items.',
)]

IDS_ONLY_OPT = Annotated[bool | None, typer.Option(
    help='Display only the IDs of the items.',
)]

FORCE_OPT = Annotated[bool | None, typer.Option(
    help='Force execution of the operation.',
)]



from sqlalchemy import ColumnElement

from reling.db.models import DialogueExam, Language, TextExam
from .modalities import Modality

__all__ = [
    'get_filter',
]


def get_filter(language: Language, modality: Modality, model: type[TextExam | DialogueExam]) -> ColumnElement[bool]:
    """Get the filtering condition for the given language, modality, and model."""
    return (model.source_language_id if modality == Modality.COMPREHENSION else model.target_language_id) == language.id
